%  
% CONSISTENCY PROTOCOL FOR DISTRIBUTED SYSTEM 
%
% Idea: si lavora in un sistema con memoria distribuita (al contrario ad es. del ticket)
% Si vuole tuttavia controllare una determinata risorsa X.
% Non si puo utilizzare un monitor centralizzato.
% Il monitor viene assegnato dinamicamente introducendo la nozione di OWNER della risorsa
% The owner funziona da monitor: colleziona gli identificatori dei processi reader
% e li passa al processo che vuole diventare writer
% Il candidato writer invalida tutti i processi e poi upgrades i suoi diritti
%
% Assunzione: invece della lista di identificatori manteniamo il NUMERO di processi reader
% tramite due variabili I,C tali che C-I=#current readers per l'owner e l'invalidator
%
% Ecco la specifica:
%
% Signature:
%
% proc(RISORSA,ACCESS_RIGHT,OWNERSHIP,INIT,CURRENT):
%  
%  ACCESS_RIGHT: NULL, READ, WRITE, INV(ALIDATE)
%  OWNERSHIP=0 (not owner) / 1 (owner)
%  IF OWNERSHIP=1 or ACCESS_RIGHT=INV --> COUNT-INIT= current number of reading processes 
%

%
% Initialization Phase: creo M processi non owner, con accesso null alla risorsa
% 

rule([init],[init(X)],{},i0).

rule([init(X)],[proc(X,null,F,I,C),init(X)],{F=0,I=C},i1).

rule([init(X)],[proc(X,null,F,I,C),init(X1)],     {X1=X+1,F=1,I=C},i2).

% Read Request
% Un processo chiede il permesso di lettura all owner
% Owner incrementa il suo contatore
% Se Owner sta scrivendo diventa un reader
%

rule([proc(X,null,F,I,C),  proc(X,null,Fo,Io,Co)],  [proc(X,read,F,I,C),  proc(X,null,Fo,Io,Co1)],
     {Co>=Io,Co1=Co+1,Fo=1},r1).

rule([proc(X,null,F,I,C),  proc(X,read,Fo,Io,Co)],  [proc(X,read,F,I,C),  proc(X,read,Fo,Io,Co1)],
     {Co>=Io,Co1=Co+1,Fo=1},r2).

rule([proc(X,null,F,I,C),  proc(X,write,Fo,Io,Co)], [proc(X,read,F,I,C),  proc(X,read, Fo,Io,Co1)],
     {Co>=Io,Co1=Co+1,Fo=1},r3).

%
% Write request
% Un processo  
%

rule([proc(X,null,F,I,C),  proc(X,Mode,Fo,Io,Co)],  [proc(X,inv, F,Io,Co),proc(X,null,Fo1,Io,Co)],
     {Fo=1,Fo1=0},r4).

rule([proc(X,inv,F,I,C),   proc(X,read,Fo,Io,Co)],  [proc(X,inv,F,I,C1),  proc(X,null,Fo,Io,Co)],
     {C>I,C1=C-1},r5).
     
rule([proc(X,inv,  F,I,C)],[proc(X,write,F1,I,C)],
     {F1=1,C=I},r6).

rule([proc(X,read,F,I,C)], [proc(X,null,F,I,C)],
     {},r7).
     
rule([proc(X,write,F,I,C)],[proc(X,read, F,I,C)],
     {},r8).

rule([proc(X,read,Fo,Io,Co)],[proc(X,inv,Fo,Io,Co)],
     {Fo=1},r9).

     
unsafe([         
	 unsafe([proc(X,_,F,_,_),proc(X,_,F,_,_)],{F=1})
]).

old_invariants([]).    

invariants([]).

autoload:-['examples/absowner.pl'].
