%  
% INVALIDATION LOOP FOR AN UNBOUNDED NUMBER OF MONITORS
% COUNTERS ARE PARAMETRIC
% 
% Signature
%
% count(ID,INIT,CURRENT):
%       conta il numero di readers associati alla risorsa ID
% 	ID=identifier, 
% 	INIT,CURRENT=Initial and current values of the counter
%
% In questo esempio i readers devono aspettare che  il writer finisca 
% di scrivere. Quando rilascia la regione critica allora i readers possono 
% entrare.
% Safety: mutua esclusione tra readers & writers
% Unsafe states:  write(ID) | count(ID,I,C): C>I (almeno un reader)
%

% Initialization Phase: M clients, counters are created dynamically 

rule([init],[init_client],{},i1).

rule([init_client],[init_client,null],{},i2).

rule([init_client],[init_count(_)],{},i3).

rule([init_monitor(X)],[idle(X,I,C),init_monitor(X1)],{X1>X,I=C},i4).

% Read request 

rule([null],[readreq(_)],{},r1).

rule([readreq(Y),idle(Y,I,C)],[idle(Y,I,C1)],{C>=I,C1=C+1},r2).

rule([readreq(Y),idle(Y,I,C)],[idle(Y,I,C1),read(Y)],{C=I,C1=C+1},r3).

rule([idle(Y,I,C)],[idle(Y,I,C1)],{C>I,C1=C-1},r4).

rule([idle(Y,I,C),read(Y)],[idle(Y,I,C1)],{C=I+1,C1=I},r5).

% Nota: monitor locked -> non si puo leggere (bisogna aspettare)

% Write request 

rule([null],[writereq(_)],{},w1).

rule([writereq(Y),idle(Y,I,C)],[invalidate(Y,I,C)],{},w2).

rule([invalidate(Y,I,C)],[invalidate(Y,I,C1),null],{C>I,C1=C-1},w3).

rule([invalidate(Y,I,C),read(Y)],[locked(Y,I,C1),write(Y)],{C=I+1,C1=I},w4).

rule([locked(Y,I,C),write(Y)],[idle(Y,I,C)],{},w5).

unsafe([unsafe([write(X),read(X)],{})]).

old_unsafe([unsafe([write(X),read(X)],{}),
        unsafe([init_monitor(_),init_monitor(_)],{}),
	unsafe([init_client,init_client],{}),
	unsafe([init,init],{}),
        unsafe([idle(C,_,_),idle(C,_,_)],{}),
	unsafe([locked(L,_,_),locked(L,_,_)],{}),
        unsafe([invalidate(I,_,_),invalidate(I,_,_)],{})
       ]).
       
invariants([
        inv([init_monitor(_),init_monitor(_)],{}),
	inv([init_client,init_client],{}),
	inv([init,init],{}),
        inv([idle(C,_,_),idle(C,_,_)],{}),
	inv([locked(L,_,_),locked(L,_,_)],{}),
        inv([invalidate(I,_,_),invalidate(I,_,_)],{})
       ]).

