%  
% CONSISTENCY PROTOCOL FOR DISTRIBUTED SYSTEM 
%
% Idea: si lavora in un sistema con memoria distribuita (al contrario ad es. del ticket)
% Si vuole tuttavia controllare una determinata risorsa.
% Non si puo utilizzare un monitor centralizzato.
% Il monitor viene assegnato dinamicamente introducendo la nozione di OWNER della risorsa
% The owner funziona da monitor: colleziona gli identificatori dei processi reader
% e li passa al processo che vuole diventare writer
% Il candidato writer invalida tutti i processi e poi upgrades i suoi diritti
%
% Assunzione: invece della lista di identificatori manteniamo il NUMERO di processi reader
% tramite due variabili I,C tali che C-I=#current readers per l'owner e l'invalidator
%
% Ecco la specifica:
%
% Signature:
%
% proc(ACCESS_RIGHT,OWNERSHIP,INIT,CURRENT):
%  
%  ACCESS_RIGHT: NULL, READ, WRITE, INV(ALIDATE)
%  OWNERSHIP=0 (not owner) / 1 (owner)
%  IF OWNERSHIP=1 or ACCESS_RIGHT=INV --> COUNT-INIT= current number of reading processes 
%

% Initialization Phase: creo M processi non owner, con accesso null alla risorsa

rule([init],[proc(null,F,I,C),init],{F=0,I=C},i1).

rule([init],[proc(null,F,I,C)],     {F=1,I=C},i2).

% Read Request
% Un processo chiede il permesso di lettura a owner
% Owner incrementa il suo contatore
% Se Owner sta scrivendo diventa un reader


rule([proc(null,F,I,C),  proc(null,Fo,Io,Co)],  [proc(null,Fo,Io,Co1)],
     {Co>=Io,Co1=Co+1,Fo=1},r1).

rule([proc(null,F,I,C),  proc(read,Fo,Io,Co)],  [proc(read,Fo,Io,Co1)],
     {Co>=Io,Co1=Co+1,Fo=1},r2).

rule([proc(null,F,I,C),  proc(write,Fo,Io,Co)], [proc(read, Fo,Io,Co1)],
     {Co>=Io,Co1=Co+1,Fo=1},r3).

% Write request

rule([proc(null,F,I,C),  proc(Mode,Fo,Io,Co)],  [proc(inv, F,Io,Co),proc(null,Fo1,Io,Co)],
     {Fo=1,Fo1=0},r4).

rule([proc(inv,F,I,C)],  [proc(inv,F,I,C1),  proc(null,Fo,Io,Co)],
     {C>I,C1=C-1,Fo=0},r5).

rule([proc(inv,  F,I,C)],[proc(write,F1,I,C)],
     {F1=1,C=I},r6).
    
rule([proc(write,F,I,C)],[proc(read, F,I,C)],
     {},r8).

rule([proc(read,Fo,Io,Co)],[proc(inv,Fo,Io,Co)],
     {Fo=1},r9).

     
unsafe([         
	 unsafe([proc(read,_,_,_),proc(write,_,_,_)],{})
]).

invariants([]).    

old_invariants([ 
       inv([proc(_,F,_,_),proc(_,F,_,_)],{F=1})
       ]).

autoload:-['examples/twoabsowner.pl'].
