:-dynamic f/3.
:-dynamic app/2.

:- prolog_flag(redefine_warnings,_,off).

:- use_module(library(terms)).
:- use_module(library(lists)).
:- use_module(library(clpr)).
:- use_module(library(tcltk)).
:- use_module(library(charsio)).
:- use_module(library(system)).

% MODULE: checker.pl
%

notallowed(M,C,Invariants):-
  member(inv(NM,NC),Invariants),
  msubsumes(M,C,NM,NC).

allowedN(M,C,Invariants):- 
 \+(notallowed(M,C,Invariants)).
  
get_constraints([],{}):-!.

get_constraints([[_]-{C}],{C}):-!.

get_constraints([[_]-{C}|R],{C,N}):-
 get_constraints(R,{N}),!.

% entails(C,D): C entails D
%  - C,D are constraints of the form {c1,...,cn}
%

entails(C,D):-
 C,entailed(D).

%

subsumed({},{}).

subsumed({C},{D}):-
 call_residue(entails({C},(D)),_).

subsumed({C},{}):-
 call_residue({C},_).

%


% factsubsumed(H,D,H1,D1): (H1,D1) subsumes (H,D)
%   - H,H1 are atoms
%   - D,D1 are constraints

factsubsumed_with_copy(Atom,Cnst,Atom1,Cnst1):-
 copy_term((Atom,Cnst),(H,D)),
 copy_term((Atom1,Cnst1),(H1,D1)),
 subsumes_chk(H1,H),
 unify_with_occurs_check(H,H1),
 simplify(D,DS),
 simplify(D1,DS1),
 subsumed(DS,DS1).  

factsubsumed(H,D,H1,C):-
 subsumes_chk(H1,H),
 unify_with_occurs_check(H,H1),
 simplify(D,D1),
 simplify(C,C1),
 subsumed(D1,C1).

% unify(C,D,R):  R is equivalent to (C and D)
%

unify({},{},{}):-!.
 
unify({},C,R):-  % pay attention call_residue({},L) fails
 call_residue(C,S),
 get_constraints(S,R),!.

unify({D},{},R):-
 call_residue({D},S),
 get_constraints(S,R),!.

unify({D},{C},R):- 
 call_residue({C,D},S),
 get_constraints(S,R),!.

%
% project(C,T,PC): PC is the projection of C over the variables of T
%
 
project({}, _, {}) :- !.

project(Cons, Term, PCons) :-
    !,
    retractall(projection_temporary(_)),
    assert((projection_temporary(Term) :- Cons)),
    call_residue(projection_temporary(Term), TmpPCons),
    get_constraints(TmpPCons, PCons).
 
 
% atom_cons_unify(H,B,B1,C,C1,D2)
%

atom_cons_unify(H,B,B1,C,C1,D2):-
 unify_with_occurs_check(B1,B),
 unify(C,C1,D),
 project(D,H,D1),
 simplify(D1,D2).
    

% satisfiable(C,D):  C and D is satisfiable
% 
 
satisfiable({},{}).

satisfiable({},{D}):-
 call_residue({D},_).
 
satisfiable({D},{}):-
 call_residue({D},_).

satisfiable({D},{C}):-
 call_residue({D},_),
 call_residue({C},_).

% simplify(C,D).
%
 
simplify({},{}):-!.
 
simplify(C,D):-
 call_residue(C,R),
 get_constraints(R,D).




init(L):-false.
    

combine({},{},{}).

combine({},{C},{C}).

combine({C},{},{C}).

combine({C},{D},{C,D}).
   
    
removeC({K>=D,W},W1):-
  nonvar(K),
  K=A-B,
  D=\=0,!,
  removeC({W},W1).    
  
removeC({K=<D,W},W1):-
  nonvar(K),
  K=A-B,
  D=\=0,!,
  removeC({W},W1). 
     
removeC({K>=D},{}):-
  nonvar(K),
  K=A-B,
  D=\=0,
  !.

removeC({K=<D},{}):-
  nonvar(K),
  K=A-B,
  D=\=0,
  !.
 
removeC({K>=D},{K>=D}):-!.

removeC({K=<D},{K=<D}):-!.

removeC({K=<D},{K=<D}):-
  nonvar(K),
  K=A-B,
  D=:=0,
  !.
 
 removeC({W,W1},W2):-
  removeC({W1},W3),
  combine({W},W3,W2),!.

removeC({},{}).
 
relaxC(X,Y,X>Y).  
%relaxC(X,Y,X>=Y).  

relaxEq({X = Y + C,W},W2):-
  var(X),
  var(Y),
  integer(C),
  C>0,!,
  relaxEq({W},W1),
  relaxC(X,Y,D),
  combine({D},W1,W2). 

relaxEq({X = Y + C,W},W2):-
  var(X),
  var(Y),
  C<0,!,
  relaxEq({W},W1),
  relaxC(Y,X,D),
  combine({D},W1,W2). 

relaxEq({X = C + Y,W},W2):-
  var(X),
  var(Y),
  C>0,!,
  relaxEq({W},W1),
  relaxC(X,Y,D),
  combine({D},W1,W2). 
  
relaxEq({X = C + Y,W},W2):-
  var(X),
  var(Y),
  C<0,!,
  relaxC(Y,X,D),
  relaxEq({W},W1),
  combine({D},W1,W2). 

relaxEq({X = Y + C},{D}):-
  var(X),
  var(Y),
  integer(C),
  C>0,!,
  relaxC(X,Y,D). 

relaxEq({X = Y + C},{D}):-
  var(X),
  var(Y),
  C<0,!,
  relaxC(Y,X,D).

relaxEq({X = C + Y},{D}):-
  var(X),
  var(Y),
  C>0,!,
  relaxC(X,Y,D). 
  
relaxEq({X = C + Y},{D}):-
  var(X),
  var(Y),
  C<0,!,
  relaxC(Y,X,D).
  
relaxEq({X = C + Y},{D}):-
  var(X),
  var(Y),
  C<0,!,
  relaxC(Y,X,D).

relaxEq({X - Y > C},{D}):-
  var(X),
  var(Y),
  C>0,!,
  relaxC(X,Y,D).

relaxEq({X - Y > C},{}):-
  var(X),
  var(Y),
  C<0,!.

relaxEq({C,W},W2):-
  relaxEq({W},W1),
  combine({C},W1,W2). 

relaxEq({C},{C}):-
  \+(C = (A,B)).

relaxEq({},{}).


widen(M,C,M,C1):-
 relaxEq(C,C1),
 storeApp(M,C,C1).

storeApp(M,C,C1):- 
 C==C1,!.

storeApp(M,C,C1):-
 assert(app(old(M,C),new(M,C1))).


mysublist([],[],[]).

mysublist([X|L],[X|M],N):-
 mysublist(L,M,N).

mysublist([X|L],M,[X|N]):-
 mysublist(L,M,N).

munify(M1,C1,M2,C2,D):-
 unify_with_occurs_check(M1,M2),
 unify(C1,C2,D).


sel(SubM,SupM,SubSupM,Rest):-
 length(SubM,LSubM),
 1 =< LSubM,
 length(SupM,LSupM),
 LSubM =< LSupM,
 sel(SubM,SupM,[],SubSupM,Rest).

sel([A|SubM],[B|RestM],SupM,[B|L],R):-
  functor(A,F,N),
  functor(B,F,N),
  append(SupM,RestM,AuxM),
  sel(SubM,AuxM,[],L,R).
 
sel([A|SubM],[B|RestM],SupM,L,R):-
  sel([A|SubM],RestM,[B|SupM],L,R).

sel([],Rest,_,[],Rest).

% msubsumes(N,C,M,D) =  den(N,C) =< den(M,D)

msubsumes(N,C,M,D):-
 sel(M,N,SN,_),
 subsumes_chk(M,SN),
 copy_term((SN,C),(CSN,CC)),
 copy_term((M, D),(CM, CD)),
 project(CC,CSN,PCC),
 unify_with_occurs_check(CSN,CM),
 simplify(PCC,SPCC),
 simplify(CD,SCD),
 subsumed(SPCC,SCD).  
 
apply_rule(M,Cm,A,B,Cab,H,Ch):-
 mysublist(M,SM,RM),
 sel(SM,B,SB,_),
 munify(SB,Cab,SM,Cm,Cu),
 append(A,RM,H),
 project(Cu,H,Cpu),
 simplify(Cpu,Ch).
 
apply_init:-
 f(_,M,_,_,_,_),
 init(M),
 write('Violation at '),
 write(M),!. 
 
apply_msubsumes(NM,NC):-
 f(_,M,C,_,_,_),
% write('^'),
% flush_output,
 msubsumes(NM,NC,M,C),!. 

 
check_asserta(f(Step,NM,NC,Nf,Pref,Clause)):-
 asserta(f(Step,NM,NC,Nf,Pref,Clause)). 
 
removeAllF(NM,NC):-
 setof(_,removeF(NM,NC),_),!.

removeAllF(_,_):-!.

removeF(NM,NC):-
 clause(f(_,M,C,_,_,_),true,Ref),
 msubsumes(M,C,NM,NC),
 erase(Ref). 
 
apply_widen(NS,NM,NC,WM,WC):-
% widen(NM,NC,WM,WC).
 NM=WM,NC=WC.
 
old_apply_widen(NS,NM,NC,NM,NC).

select_and_apply(CS,NS):-
 invariants(Invariants),
 f(CS,M,Cm,Nf,_,_),
 copy_term((M,Cm),(CM,CCm)),
 rule(A,B,Cab,RuleNo),
 apply_rule(M,Cm,A,B,Cab,NM,NC),
 allowedN(NM,NC,Invariants),
 \+(msubsumes(NM,NC,CM,CCm)),
 \+(apply_msubsumes(NM,NC)), 
 apply_widen(NS,NM,NC,WM,WC),
% removeAllF(WM,WC),
 retract(countF(CountF)),
 CountF1 is CountF+1,
 assert(countF(CountF1)), 
 asserta(f(NS,WM,WC,CountF,Nf,RuleNo)),
 write('.'),
 flush_output.

loop(N):-
 N1 is N+1,
 nl,
 write('Step'),
 write(N1),
 nl,
 flush_output,
 setof(_,select_and_apply(N,N1),_),
 safe,!,
 loop(N1).

loop(N):- 
 safe,
 nl,
 write('--------------------------------------------------------'),nl,
 write('The system is safe :)'),nl,
 write('The fixpoint *without* the atom -init- has been reached!'),nl,
 write('--------------------------------------------------------').

loop(N):- 
 nl,
 write('----------------------------------'),nl,
 write('The system is unsafe :('),nl,
 write('The atom -init- has been detected!'),nl,
 write('----------------------------------').

go:-
 unsafe(L),
 clearF(L),
 clearApp,
 loop(0).

clearF(L):-
 retractall(f(_,_,_,_,_,_)),
 retractall(countF(_)),
 assert(countF(1)),
 assertList(L).
 
clearApp:-
 retractall(app(_,_)).
 
assertList([unsafe(Unsafe,C)|Rest]):-
 !, 
 retract(countF(CountF)),
 CountF1 is CountF+1,
 assert(countF(CountF1)),
 assert(f(0,Unsafe,C,CountF,0,0)),
 assertList(Rest).
  
assertList([]). 

safe:- 
 \+ f(_,[init],_,_,_,_).

listF:-listing(f).

listR:-listing(rule).

listU:-listing(unsafe).

listA:-listing(app).

loadms:-
['examples/multiserverticket.pl'].

ldss:-
['examples/singleserverticket.pl'].

ldil:-
['examples/invloop.pl'].

ldinv:-
['examples/invalidation.pl'].

ldinv1:-
['examples/invalidation1.pl'].

ldinv2:-
['examples/invalidation2.pl'].

ldrace:-
['examples/invalidationrace.pl'].

ldlead:-
['examples/leader.pl'].

ldiw:-
['examples/inv_no_preemption_writers.pl'].

ldow:-
['examples/absowner.pl'].

ldabo:-
['examples/twoabsowner.pl'].

examples:-
 write('load an example and then type go'),nl,nl,
 write('examples/invalidation.pl       --> command: loadinv'),nl,
 write('examples/inv_no_preemption_writers.pl   --> command: loadiw'),nl,
 write('examples/invalidation1.pl      --> command: loadinv1'),nl,
 write('examples/invalidation2.pl      --> command: loadinv2'),nl,
 write('examples/invalidationrace.pl   --> command: loadrace (solo inval./senza
 loop)'),nl,
 write('examples/multiserverticket.pl  --> command: loadms'),nl,
 write('examples/singleserverticket.pl --> command: loadss'),nl,
 write('examples/invloop               --> command: loadil'),nl,
 write('examples/leader.pl             --> command: loadlead').
 
attack:-
 f(Step, [init], Cstr, K, Prev, Rule),
 tracer(Step,K).
 
tracer(Step,K):-
 Step>=0,!,
 f(Step, State, Cstr, K, Prev, Rule),
 write('Step'), write(Step),nl,
 write(State),nl,
 write(Cstr),nl,
 rule(H,B,C,Rule), 
 write(H),write('-->'),write(B),nl,
 %read(_),
 Step1 is Step-1,
 tracer(Step1,Prev).

tracer(Step,K).

ls:-listing(f/6).
lr:-listing(rule/4).
lz:-listing(z/2).
lc:-listing(count/1).
lf:-listing(info/4).
